@@ -9,24 +9,24 @@
 
 # Source-form features:
 # ---------------------
-# 
+#
 # Bullet \b
 #   Bullets the paragraph. Rest of paragraph is indented to cope. In
 #   HTML, consecutive groups of bulleted paragraphs become unordered
 #   lists.
-# 
+#
 # Emphasis \e{foobar}
 #   produces `_foobar_' in text and italics in HTML, PS, RTF
-# 
+#
 # Inline code \c{foobar}
 #   produces ``foobar'' in text, and fixed-pitch font in HTML, PS, RTF
-# 
+#
 # Display code
 # \c  line one
 # \c   line two
 #   produces fixed-pitch font where appropriate, and doesn't break
 #   pages except sufficiently far into the middle of a display.
-# 
+#
 # Chapter, header and subheader
 # \C{intro} Introduction
 # \H{whatsnasm} What is NASM?
@@ -40,17 +40,17 @@
 #   Expands to `Chapter 1', `Section 1.1', `Section 1.1.1'. \K has an
 #   initial capital whereas \k doesn't. In HTML, will produce
 #   hyperlinks.
-# 
+#
 # Web link \W{http://foobar/}{text} or \W{mailto:me@here}\c{me@here}
 #   the \W prefix is ignored except in HTML; in HTML the last part
 #   becomes a hyperlink to the first part.
-# 
+#
 # Literals \{ \} \\
 #   In case it's necessary, they expand to the real versions.
-# 
+#
 # Nonbreaking hyphen \-
 #   Need more be said?
-# 
+#
 # Source comment \#
 #   Causes everything after it on the line to be ignored by the
 #   source-form processor.
@@ -79,9 +79,17 @@
 #   defines document metadata, such as authorship, title and copyright;
 #   different output formats use this differently.
 #
+# Include subfile
+# \&{filename}
+#  Includes filename. Recursion is allowed.
+#
+ 
+use IO::File;
 
 $diag = 1, shift @ARGV if $ARGV[0] eq "-d";
 
+($out_format) = @ARGV;
+
 $| = 1;
 
 $tstruct_previtem = $node = "Top";
@@ -95,16 +103,8 @@ print "Reading input...";
 $pname = "para000000";
 @pnames = @pflags = ();
 $para = undef;
-while (<>) {
-  chomp;
-  if (!/\S/ || /^\\(IA|IR|M)/) { # special case: \IA \IR \M imply new-paragraph
-    &got_para($para);
-    $para = undef;
-  }
-  if (/\S/) {
-    s/\\#.*$//; # strip comments
-    $para .= " " . $_;
-  }
+while (defined($_ = <STDIN>)) {
+  &check_include($_);
 }
 &got_para($para);
 print "done.\n";
@@ -127,22 +127,57 @@ if ($diag) {
 }
 
 # OK. Write out the various output files.
-print "Producing text output: ";
-&write_txt;
-print "done.\n";
-print "Producing HTML output: ";
-&write_html;
-print "done.\n";
-print "Producing Texinfo output: ";
-&write_texi;
-print "done.\n";
-print "Producing WinHelp output: ";
-&write_hlp;
-print "done.\n";
-print "Producing Documentation Intermediate Paragraphs: ";
-&write_dip;
-print "done.\n";
+if ($out_format eq 'txt') {
+    print "Producing text output: ";
+    &write_txt;
+    print "done.\n";
+} elsif ($out_format eq 'html') {
+    print "Producing HTML output: ";
+    &write_html;
+    print "done.\n";
+} elsif ($out_format eq 'texi') {
+    print "Producing Texinfo output: ";
+    &write_texi;
+    print "done.\n";
+} elsif ($out_format eq 'hlp') {
+    print "Producing WinHelp output: ";
+    &write_hlp;
+    print "done.\n";
+} elsif ($out_format eq 'dip') {
+    print "Producing Documentation Intermediate Paragraphs: ";
+    &write_dip;
+    print "done.\n";
+} else {
+    die "$0: unknown output format: $out_format\n";
+}
 
+sub check_include {
+  local $_ = shift;
+  if (/\\& (\S+)/) {
+     &include($1);
+  } else {
+     &get_para($_);
+  }
+}
+sub get_para($_) {
+  chomp;
+  if (!/\S/ || /^\\(IA|IR|M)/) { # special case: \IA \IR \M imply new-paragraph
+    &got_para($para);
+    $para = undef;
+  }
+  if (/\S/) {
+    s/\\#.*$//; # strip comments
+    $para .= " " . $_;
+  }
+}
+sub include {
+  my $name = shift;
+  my $F = IO::File->new($name)
+     or die "Cannot open $name: $!";
+  while (<$F>) {
+     &check_include($_);
+  }
+}
 sub got_para {
   local ($_) = @_;
   my $pflags = "", $i, $w, $l, $t;
@@ -401,7 +436,7 @@ sub indexsort {
            } map { # transform array into list of 3-element lists
 	     my $ientry = $idxmap{$_};
 	     my $a = substr($$ientry[0],2);
-	     $a =~ tr/A-Za-z//cd;
+	     $a =~ tr/A-Za-z0-9//cd;
 	     [$_, uc($a), substr($$ientry[0],0,2)]
 	   } keys %idxmap;
 
@@ -458,7 +493,7 @@ sub fixup_xrefs {
       if ($$pname[$i] =~ /^k/) {
         $k = $$pname[$i];
         $caps = ($k =~ /^kK/);
-	$k = substr($k,2);	
+	$k = substr($k,2);
         $repl = $refs{$k};
 	die "undefined keyword `$k'\n" unless $repl;
 	substr($repl,0,1) =~ tr/a-z/A-Z/ if $caps;
@@ -1255,11 +1290,12 @@ sub write_hlp {
       # Code paragraph.
       print "\\keep\\f1\\sb120\n";
       foreach $i (@$pname) {
+	my $x = $i;
         warn "code line longer than 68 chars: $i\n" if length $i > 68;
-	$i =~ s/\\/\\\\/g;
-	$i =~ s/\{/\\\{/g;
-	$i =~ s/\}/\\\}/g;
-        print "$i\\par\\sb0\n";
+	$x =~ s/\\/\\\\/g;
+	$x =~ s/\{/\\\{/g;
+	$x =~ s/\}/\\\}/g;
+        print "$x\\par\\sb0\n";
       }
       $newpar = "\\pard\\f0\\sb120\n";
     } elsif ($ptype eq "bull" || $ptype eq "norm") {
@@ -1417,4 +1453,3 @@ sub write_dip {
   }
   close(PARAS);
 }
-
