@@ -38,7 +38,7 @@ struct forwrefinfo {            /* info held on forward refs. */
 static int get_bits(char *value);
 static uint32_t get_cpu(char *cpu_str);
 static void parse_cmdline(int, char **);
-static void assemble_file(char *);
+static void assemble_file(char *, StrList **);
 static void register_output_formats(void);
 static void report_error_gnu(int severity, const char *fmt, ...);
 static void report_error_vc(int severity, const char *fmt, ...);
@@ -88,9 +88,13 @@ enum op_type {
     op_normal,                  /* Preprocess and assemble */
     op_preprocess,              /* Preprocess only */
     op_depend,                  /* Generate dependencies */
-    op_depend_missing_ok,	/* Generate dependencies, missing OK */
 };
 static enum op_type operating_mode;
+/* Dependency flags */
+static bool depend_emit_phony = false;
+static bool depend_missing_ok = false;
+static const char *depend_target = NULL;
+static const char *depend_file = NULL;
 
 /*
  * Which of the suppressible warnings are suppressed. Entry zero
@@ -133,7 +137,7 @@ static const char *suppressed_what[ERR_WARN_MAX+1] = {
  * not preprocess their source file.
  */
 
-static void no_pp_reset(char *, int, efunc, evalfunc, ListGen *);
+static void no_pp_reset(char *, int, efunc, evalfunc, ListGen *, StrList **);
 static char *no_pp_getline(void);
 static void no_pp_cleanup(int);
 static Preproc no_pp = {
@@ -216,7 +220,7 @@ static void define_macros_early(void)
 	strftime(temp, sizeof temp, "__UTC_TIME_NUM__=%H%M%S", &gm);
 	pp_pre_define(temp);
     }
-    
+
     if (gm_p)
 	posix_time = posix_mktime(&gm);
     else if (lt_p)
@@ -239,8 +243,51 @@ static void define_macros_late(void)
     pp_pre_define(temp);
 }
 
+static void emit_dependencies(StrList *list)
+{
+    FILE *deps;
+    int linepos, len;
+    StrList *l, *nl;
+
+    if (depend_file && strcmp(depend_file, "-")) {
+	deps = fopen(depend_file, "w");
+	if (!deps) {
+	    report_error(ERR_NONFATAL|ERR_NOFILE|ERR_USAGE,
+			 "unable to write dependency file `%s'", depend_file);
+	    return;
+	}
+    } else {
+	deps = stdout;
+    }
+    
+    linepos = fprintf(deps, "%s:", depend_target);
+    for (l = list; l; l = l->next) {
+	len = strlen(l->str);
+	if (linepos + len > 62) {
+	    fprintf(deps, " \\\n ");
+	    linepos = 1;
+	}
+	fprintf(deps, " %s", l->str);
+	linepos += len+1;
+    }
+    fprintf(deps, "\n\n");
+    
+    for (l = list; l; l = nl) {
+	if (depend_emit_phony)
+	    fprintf(deps, "%s:\n\n", l->str);
+	    
+	nl = l->next;
+	nasm_free(l);
+    }
+
+    if (deps != stdout)
+	fclose(deps);
+}
+
 int main(int argc, char **argv)
 {
+    StrList *depend_list = NULL, **depend_ptr;
+
     time(&official_compile_time);
 
     pass0 = 1;
@@ -284,22 +331,27 @@ int main(int argc, char **argv)
     /* define some macros dependent of command-line */
     define_macros_late();
 
+    depend_ptr = (depend_file || (operating_mode == op_depend))
+	? &depend_list : NULL;
+    if (!depend_target)
+	depend_target = outname;
+
     switch (operating_mode) {
-    case op_depend_missing_ok:
-	pp_include_path(NULL);	/* "assume generated" */
-	/* fall through */
     case op_depend:
         {
             char *line;
-            preproc->reset(inname, 0, report_error, evaluate, &nasmlist);
+
+	    if (depend_missing_ok)
+		pp_include_path(NULL);	/* "assume generated" */
+
+            preproc->reset(inname, 0, report_error, evaluate, &nasmlist,
+			   depend_ptr);
             if (outname[0] == '\0')
                 ofmt->filename(inname, outname, report_error);
             ofile = NULL;
-            fprintf(stdout, "%s: %s", outname, inname);
             while ((line = preproc->getline()))
                 nasm_free(line);
             preproc->cleanup(0);
-            putc('\n', stdout);
         }
         break;
 
@@ -321,8 +373,10 @@ int main(int argc, char **argv)
 
             location.known = false;
 
-/*      pass = 1; */
-            preproc->reset(inname, 2, report_error, evaluate, &nasmlist);
+	    /* pass = 1; */
+            preproc->reset(inname, 2, report_error, evaluate, &nasmlist,
+			   depend_ptr);
+
             while ((line = preproc->getline())) {
                 /*
                  * We generate %line directives if needed for later programs
@@ -378,7 +432,7 @@ int main(int argc, char **argv)
 
             ofmt->init(ofile, report_error, define_label, evaluate);
 
-            assemble_file(inname);
+            assemble_file(inname, depend_ptr);
 
             if (!terminate_after_phase) {
                 ofmt->cleanup(using_debug_info);
@@ -398,6 +452,9 @@ int main(int argc, char **argv)
         break;
     }
 
+    if (depend_list)
+	emit_dependencies(depend_list);
+
     if (want_usage)
         usage();
 
@@ -448,6 +505,88 @@ static void copy_filename(char *dst, const char *src)
     strncpy(dst, src, FILENAME_MAX);
 }
 
+/*
+ * Convert a string to Make-safe form
+ */
+static char *quote_for_make(const char *str)
+{
+    const char *p;
+    char *os, *q;
+
+    size_t n = 1;		/* Terminating zero */
+    size_t nbs = 0;
+
+    if (!str)
+	return NULL;
+
+    for (p = str; *p; p++) {
+	switch (*p) {
+	case ' ':
+	case '\t':
+	    /* Convert N backslashes + ws -> 2N+1 backslashes + ws */
+	    n += nbs + 2;
+	    nbs = 0;
+	    break;
+	case '$':
+	case '#':
+	    nbs = 0;
+	    n += 2;
+	    break;
+	case '\\':
+	    nbs++;
+	    n++;
+	    break;
+	default:
+	    nbs = 0;
+	    n++;
+	break;
+	}
+    }
+
+    /* Convert N backslashes at the end of filename to 2N backslashes */
+    if (nbs)
+	n += nbs;
+
+    os = q = nasm_malloc(n);
+
+    nbs = 0;
+    for (p = str; *p; p++) {
+	switch (*p) {
+	case ' ':
+	case '\t':
+	    while (nbs--)
+		*q++ = '\\';
+	    *q++ = '\\';
+	    *q++ = *p;
+	    break;
+	case '$':
+	    *q++ = *p;
+	    *q++ = *p;
+	    nbs = 0;
+	    break;
+	case '#':
+	    *q++ = '\\';
+	    *q++ = *p;
+	    nbs = 0;
+	    break;
+	case '\\':
+	    *q++ = *p;
+	    nbs++;
+	    break;
+	default:
+	    *q++ = *p;
+	    nbs = 0;
+	break;
+	}
+    }
+    while (nbs--)
+	*q++ = '\\';
+
+    *q = '\0';
+
+    return os;
+}
+
 struct textargs {
     const char *label;
     int value;
@@ -502,7 +641,7 @@ static bool process_arg(char *p, char *q)
 	case 'O':		/* Optimization level */
 	{
 	    int opt;
-	    
+
 	    if (!*param) {
 		/* Naked -O == -Ox */
 		optimizing = INT_MAX >> 1; /* Almost unlimited */
@@ -512,7 +651,7 @@ static bool process_arg(char *p, char *q)
 		    case '0': case '1': case '2': case '3': case '4':
 		    case '5': case '6': case '7': case '8': case '9':
 			opt = strtoul(param, &param, 10);
-			
+
 			/* -O0 -> optimizing == -1, 0.98 behaviour */
 			/* -O1 -> optimizing == 0, 0.98.09 behaviour */
 			if (opt < 2)
@@ -520,18 +659,18 @@ static bool process_arg(char *p, char *q)
 			else
 			    optimizing = opt;
 			break;
-			
+
 		    case 'v':
 		    case '+':
 			param++;
 			opt_verbose_info = true;
 			break;
-			
+
 		    case 'x':
 			param++;
 			optimizing = INT_MAX >> 1; /* Almost unlimited */
 			break;
-			
+
 		    default:
 			report_error(ERR_FATAL,
 				     "unknown optimization option -O%c\n",
@@ -579,6 +718,7 @@ static bool process_arg(char *p, char *q)
 			     " output format `%s'",
 			     param, ofmt->shortname);
 	    }
+	    using_debug_info = true;
 	    break;
 
 	case 'X':		/* specify error reporting format */
@@ -709,7 +849,39 @@ static bool process_arg(char *p, char *q)
             break;
 
         case 'M':
-            operating_mode = p[2] == 'G' ? op_depend_missing_ok : op_depend;
+	    switch (p[2]) {
+	    case 0:
+		operating_mode = op_depend;
+		break;
+	    case 'G':
+		operating_mode = op_depend;
+		depend_missing_ok = true;
+		break;
+	    case 'P':
+		depend_emit_phony = true;
+		break;
+	    case 'D':
+		depend_file = q;
+		advance = true;
+		break;
+	    case 'T':
+		depend_target = q;
+		advance = true;
+		break;
+	    case 'Q':
+		depend_target = quote_for_make(q);
+		advance = true;
+		break;
+	    default:
+		report_error(ERR_NONFATAL|ERR_NOFILE|ERR_USAGE,
+			     "unknown dependency option `-M%c'", p[2]);
+		break;
+	    }
+	    if (advance && (!q || !q[0])) {
+		report_error(ERR_NONFATAL|ERR_NOFILE|ERR_USAGE,
+			     "option `-M%c' requires a parameter", p[2]);
+		break;
+	    }
             break;
 
         case '-':
@@ -941,8 +1113,10 @@ static void parse_cmdline(int argc, char **argv)
     if (!*inname)
         report_error(ERR_NONFATAL | ERR_NOFILE | ERR_USAGE,
                      "no input file specified");
-    else if (!strcmp(inname, errname) || !strcmp(inname, outname) ||
-	!strcmp(inname, listname))
+    else if (!strcmp(inname, errname) ||
+	     !strcmp(inname, outname) ||
+	     !strcmp(inname, listname) ||
+	     (depend_file && !strcmp(inname, depend_file)))
 	report_error(ERR_FATAL | ERR_NOFILE | ERR_USAGE,
 		     "file `%s' is both input and output file",
 		     inname);
@@ -969,7 +1143,7 @@ static const char *directives[] = {
 };
 static enum directives getkw(char **directive, char **value);
 
-static void assemble_file(char *fname)
+static void assemble_file(char *fname, StrList **depend_ptr)
 {
     char *directive, *value, *p, *q, *special, *line, debugid[80];
     insn output_ins;
@@ -1013,7 +1187,9 @@ static void assemble_file(char *fname)
             raa_free(offsets);
             offsets = raa_init();
         }
-        preproc->reset(fname, pass1, report_error, evaluate, &nasmlist);
+        preproc->reset(fname, pass1, report_error, evaluate, &nasmlist,
+		       pass1 == 2 ? depend_ptr : NULL);
+	    
         globallineno = 0;
         if (passn == 1)
             location.known = true;
@@ -1486,6 +1662,16 @@ static void assemble_file(char *fname)
                                     TYS_ELEMENTS(output_ins.oprs[0].
                                                  offset) | TY_TBYTE;
                                 break;
+                            case I_RESO:
+                                typeinfo =
+                                    TYS_ELEMENTS(output_ins.oprs[0].
+                                                 offset) | TY_OWORD;
+                                break;
+                            case I_RESY:
+                                typeinfo =
+                                    TYS_ELEMENTS(output_ins.oprs[0].
+                                                 offset) | TY_YWORD;
+                                break;
                             case I_DB:
                                 typeinfo |= TY_BYTE;
                                 break;
@@ -1507,6 +1693,9 @@ static void assemble_file(char *fname)
 			    case I_DO:
 				typeinfo |= TY_OWORD;
 				break;
+			    case I_DY:
+				typeinfo |= TY_YWORD;
+				break;
                             default:
                                 typeinfo = TY_LABEL;
 
@@ -1789,7 +1978,7 @@ static ListGen *no_pp_list;
 static int32_t no_pp_lineinc;
 
 static void no_pp_reset(char *file, int pass, efunc error, evalfunc eval,
-                        ListGen * listgen)
+                        ListGen * listgen, StrList **deplist)
 {
     src_set_fname(nasm_strdup(file));
     src_set_linnum(0);
@@ -1802,6 +1991,13 @@ static void no_pp_reset(char *file, int pass, efunc error, evalfunc eval,
     no_pp_list = listgen;
     (void)pass;                 /* placate compilers */
     (void)eval;                 /* placate compilers */
+
+    if (deplist) {
+	StrList *sl = nasm_malloc(strlen(file)+1+sizeof sl->next);
+	sl->next = NULL;
+	strcpy(sl->str, file);
+	*deplist = sl;
+    }
 }
 
 static char *no_pp_getline(void)
