@@ -2,7 +2,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  */
 
@@ -96,10 +96,10 @@ void nasm_free(void *q)
 #endif
 {
     if (q) {
-        free(q);
 #ifdef LOGALLOC
         fprintf(logfp, "%s %d free(%p)\n", file, line, q);
 #endif
+        free(q);
     }
 }
 
@@ -149,31 +149,58 @@ char *nasm_strndup(char *s, size_t len)
 #ifndef nasm_stricmp
 int nasm_stricmp(const char *s1, const char *s2)
 {
-    while (*s1 && tolower(*s1) == tolower(*s2))
-        s1++, s2++;
-    if (!*s1 && !*s2)
-        return 0;
-    else if (tolower(*s1) < tolower(*s2))
-        return -1;
-    else
-        return 1;
+    unsigned char c1, c2;
+    int d;
+
+    while (1) {
+	c1 = *s1++;
+	c2 = *s2++;
+	d = c1-c2;
+
+	if (d)
+	    return d;
+	if (!c1)
+	    break;
+    }
+    return 0;
 }
 #endif
 
 #ifndef nasm_strnicmp
-int nasm_strnicmp(const char *s1, const char *s2, int n)
+int nasm_strnicmp(const char *s1, const char *s2, size_t n)
 {
-    while (n > 0 && *s1 && tolower(*s1) == tolower(*s2))
-        s1++, s2++, n--;
-    if ((!*s1 && !*s2) || n == 0)
-        return 0;
-    else if (tolower(*s1) < tolower(*s2))
-        return -1;
-    else
-        return 1;
+    unsigned char c1, c2;
+    int d;
+
+    while (n--) {
+	c1 = *s1++;
+	c2 = *s2++;
+	d = c1-c2;
+
+	if (d)
+	    return d;
+	if (!c1)
+	    break;
+    }
+    return 0;
 }
 #endif
 
+int nasm_memicmp(const char *s1, const char *s2, size_t n)
+{
+    unsigned char c1, c2;
+    int d;
+
+    while (n--) {
+	c1 = tolower(*s1++);
+	c2 = tolower(*s2++);
+	d = c1-c2;
+	if (d)
+	    return d;
+    }
+    return 0;
+}
+
 #ifndef nasm_strsep
 char *nasm_strsep(char **stringp, const char *delim)
 {
@@ -316,7 +343,7 @@ int64_t readnum(char *str, bool *error)
 
     if (warn)
         nasm_malloc_error(ERR_WARNING | ERR_PASS1 | ERR_WARN_NOV,
-                          "numeric constant %s does not fit in 32 bits",
+                          "numeric constant %s does not fit in 64 bits",
                           str);
 
     return result * sign;
@@ -445,7 +472,7 @@ void standard_extension(char *inname, char *outname, char *extension,
 #define LEAFSIZ (sizeof(RAA)-sizeof(RAA_UNION)+sizeof(RAA_LEAF))
 #define BRANCHSIZ (sizeof(RAA)-sizeof(RAA_UNION)+sizeof(RAA_BRANCH))
 
-#define LAYERSIZ(r) ( (r)->layers==0 ? RAA_BLKSIZE : RAA_LAYERSIZE )
+#define LAYERSHIFT(r) ( (r)->layers==0 ? RAA_BLKSHIFT : RAA_LAYERSHIFT )
 
 static struct RAA *real_raa_init(int layers)
 {
@@ -454,15 +481,13 @@ static struct RAA *real_raa_init(int layers)
 
     if (layers == 0) {
         r = nasm_zalloc(LEAFSIZ);
-        r->stepsize = 1L;
+        r->shift = 0;
     } else {
         r = nasm_malloc(BRANCHSIZ);
         r->layers = layers;
         for (i = 0; i < RAA_LAYERSIZE; i++)
             r->u.b.data[i] = NULL;
-        r->stepsize = RAA_BLKSIZE;
-        while (--layers)
-            r->stepsize *= RAA_LAYERSIZE;
+        r->shift = (RAA_BLKSHIFT-RAA_LAYERSHIFT) + layers*RAA_LAYERSHIFT;
     }
     return r;
 }
@@ -485,13 +510,12 @@ void raa_free(struct RAA *r)
 
 int64_t raa_read(struct RAA *r, int32_t posn)
 {
-    if (posn >= r->stepsize * LAYERSIZ(r))
+    if ((uint32_t)posn >= (UINT32_C(1) << (r->shift + LAYERSHIFT(r))))
         return 0;               /* Return 0 for undefined entries */
     while (r->layers > 0) {
-        ldiv_t l;
-        l = ldiv(posn, r->stepsize);
-        r = r->u.b.data[l.quot];
-        posn = l.rem;
+	int32_t l = posn >> r->shift;
+	posn &= (UINT32_C(1) << r->shift)-1;
+        r = r->u.b.data[l];
         if (!r)
             return 0;           /* Return 0 for undefined entries */
     }
@@ -505,7 +529,7 @@ struct RAA *raa_write(struct RAA *r, int32_t posn, int64_t value)
     if (posn < 0)
         nasm_malloc_error(ERR_PANIC, "negative position in raa_write");
 
-    while (r->stepsize * LAYERSIZ(r) <= posn) {
+    while ((UINT32_C(1) << (r->shift+LAYERSHIFT(r))) <= (uint32_t)posn) {
         /*
          * Must add a layer.
          */
@@ -516,7 +540,7 @@ struct RAA *raa_write(struct RAA *r, int32_t posn, int64_t value)
         for (i = 0; i < RAA_LAYERSIZE; i++)
             s->u.b.data[i] = NULL;
         s->layers = r->layers + 1;
-        s->stepsize = LAYERSIZ(r) * r->stepsize;
+        s->shift = LAYERSHIFT(r) + r->shift;
         s->u.b.data[0] = r;
         r = s;
     }
@@ -524,14 +548,13 @@ struct RAA *raa_write(struct RAA *r, int32_t posn, int64_t value)
     result = r;
 
     while (r->layers > 0) {
-        ldiv_t l;
         struct RAA **s;
-        l = ldiv(posn, r->stepsize);
-        s = &r->u.b.data[l.quot];
+	int32_t l = posn >> r->shift;
+	posn &= (UINT32_C(1) << r->shift)-1;
+        s = &r->u.b.data[l];
         if (!*s)
             *s = real_raa_init(r->layers - 1);
         r = *s;
-        posn = l.rem;
     }
 
     r->u.l.data[posn] = value;
@@ -656,6 +679,61 @@ void saa_wbytes(struct SAA *s, const void *data, size_t len)
     }
 }
 
+/* write unsigned LEB128 value to SAA */
+void saa_wleb128u(struct SAA *psaa, int value)
+{
+  char temp[64], *ptemp;
+  uint8_t byte;
+  int len;
+
+  ptemp = temp;
+  len = 0;
+  do
+  {
+     byte = value & 127;
+     value >>= 7;
+     if (value != 0) /* more bytes to come */
+        byte |= 0x80;
+     *ptemp = byte;
+     ptemp++;
+     len++;
+  } while (value != 0);
+  saa_wbytes(psaa, temp, len);
+}
+
+/* write signed LEB128 value to SAA */
+void saa_wleb128s(struct SAA *psaa, int value)
+{
+  char temp[64], *ptemp;
+  uint8_t byte;
+  bool more, negative;
+  int size, len;
+
+  ptemp = temp;
+  more = 1;
+  negative = (value < 0);
+  size = sizeof(int) * 8;
+  len = 0;
+  while(more)
+  {
+    byte = value & 0x7f;
+    value >>= 7;
+    if (negative)
+     /* sign extend */
+     value |= - (1 <<(size - 7));
+    /* sign bit of byte is second high order bit (0x40) */
+    if ((value == 0 && ! (byte & 0x40)) ||
+       ((value == -1) && (byte & 0x40)))
+       more = 0;
+    else
+      byte |= 0x80;
+    *ptemp = byte;
+    ptemp++;
+    len++;
+  }
+  saa_wbytes(psaa, temp, len);
+}
+
 void saa_rewind(struct SAA *s)
 {
     s->rblk = s->blk_ptrs;
@@ -879,24 +957,6 @@ int src_get(int32_t *xline, char **xname)
     return 0;
 }
 
-void nasm_quote(char **str)
-{
-    int ln = strlen(*str);
-    char q = (*str)[0];
-    char *p;
-    if (ln > 1 && (*str)[ln - 1] == q && (q == '"' || q == '\''))
-        return;
-    q = '"';
-    if (strchr(*str, q))
-        q = '\'';
-    p = nasm_malloc(ln + 3);
-    strcpy(p + 1, *str);
-    nasm_free(*str);
-    p[ln + 1] = p[0] = q;
-    p[ln + 2] = 0;
-    *str = p;
-}
-
 char *nasm_strcat(char *one, char *two)
 {
     char *rslt;
