@@ -3,14 +3,17 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  */
 
+#include "compiler.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
+#include <inttypes.h>
 
 #include "nasm.h"
 #include "nasmlib.h"
@@ -21,17 +24,17 @@
 struct Piece {
     struct Piece *next;
     int type;                   /* 0 = absolute, 1 = seg, 2 = sym */
-    long offset;                /* relative offset */
-    int number;                 /* symbol/segment number (4=bss) */
-    long bytes;                 /* size of reloc or of absolute data */
-    int relative;               /* TRUE or FALSE */
+    int32_t offset;		/* relative offset */
+    int number;			/* symbol/segment number (4=bss) */
+    int32_t bytes;	        /* size of reloc or of absolute data */
+    bool relative;		/* relative address? */
 };
 
 struct Symbol {
-    long strpos;                /* string table position of name */
+    int32_t strpos;		/* string table position of name */
     int flags;                  /* symbol flags */
     int segment;                /* 4=bss at this point */
-    long value;                 /* address, or COMMON variable size */
+    int32_t value;		/* address, or COMMON variable size */
 };
 
 /*
@@ -51,24 +54,24 @@ struct Symbol {
 
 struct Section {
     struct SAA *data;
-    unsigned long datalen, size, len;
-    long index;
+    uint32_t datalen, size, len;
+    int32_t index;
     struct Piece *head, *last, **tail;
 };
 
 static char as86_module[FILENAME_MAX];
 
 static struct Section stext, sdata;
-static unsigned long bsslen;
-static long bssindex;
+static uint32_t bsslen;
+static int32_t bssindex;
 
 static struct SAA *syms;
-static unsigned long nsyms;
+static uint32_t nsyms;
 
 static struct RAA *bsym;
 
 static struct SAA *strs;
-static unsigned long strslen;
+static uint32_t strslen;
 
 static int as86_reloc_size;
 
@@ -78,14 +81,15 @@ static efunc error;
 static void as86_write(void);
 static void as86_write_section(struct Section *, int);
 static int as86_add_string(char *name);
-static void as86_sect_write(struct Section *, const unsigned char *,
-                            unsigned long);
+static void as86_sect_write(struct Section *, const uint8_t *,
+                            uint32_t);
 
 static void as86_init(FILE * fp, efunc errfunc, ldfunc ldef, evalfunc eval)
 {
     as86fp = fp;
     error = errfunc;
     (void)ldef;                 /* placate optimisers */
+    (void)eval;
     stext.data = saa_init(1L);
     stext.datalen = 0L;
     stext.head = stext.last = NULL;
@@ -100,7 +104,7 @@ static void as86_init(FILE * fp, efunc errfunc, ldfunc ldef, evalfunc eval)
     stext.index = seg_alloc();
     sdata.index = seg_alloc();
     bssindex = seg_alloc();
-    syms = saa_init((long)sizeof(struct Symbol));
+    syms = saa_init((int32_t)sizeof(struct Symbol));
     nsyms = 0;
     bsym = raa_init();
     strs = saa_init(1L);
@@ -134,8 +138,11 @@ static void as86_cleanup(int debuginfo)
     saa_free(strs);
 }
 
-static long as86_section_names(char *name, int pass, int *bits)
+static int32_t as86_section_names(char *name, int pass, int *bits)
 {
+
+    (void)pass;
+
     /*
      * Default is 16 bits.
      */
@@ -160,13 +167,13 @@ static int as86_add_string(char *name)
     int pos = strslen;
     int length = strlen(name);
 
-    saa_wbytes(strs, name, (long)(length + 1));
+    saa_wbytes(strs, name, (int32_t)(length + 1));
     strslen += 1 + length;
 
     return pos;
 }
 
-static void as86_deflabel(char *name, long segment, long offset,
+static void as86_deflabel(char *name, int32_t segment, int64_t offset,
                           int is_global, char *special)
 {
     struct Symbol *sym;
@@ -216,8 +223,8 @@ static void as86_deflabel(char *name, long segment, long offset,
     nsyms++;
 }
 
-static void as86_add_piece(struct Section *sect, int type, long offset,
-                           long segment, long bytes, int relative)
+static void as86_add_piece(struct Section *sect, int type, int32_t offset,
+                           int32_t segment, int32_t bytes, int relative)
 {
     struct Piece *p;
 
@@ -247,21 +254,19 @@ static void as86_add_piece(struct Section *sect, int type, long offset,
         p->number = raa_read(bsym, segment), p->type = 2;
 }
 
-static void as86_out(long segto, const void *data, unsigned long type,
-                     long segment, long wrt)
+static void as86_out(int32_t segto, const void *data,
+		     enum out_type type, uint64_t size,
+                     int32_t segment, int32_t wrt)
 {
     struct Section *s;
-    long realbytes = type & OUT_SIZMASK;
-    long offset;
-    unsigned char mydata[4], *p;
+    int32_t offset;
+    uint8_t mydata[4], *p;
 
     if (wrt != NO_SEG) {
         wrt = NO_SEG;           /* continue to do _something_ */
         error(ERR_NONFATAL, "WRT not supported by as86 output format");
     }
 
-    type &= OUT_TYPMASK;
-
     /*
      * handle absolute-assembly (structure definitions)
      */
@@ -285,44 +290,44 @@ static void as86_out(long segto, const void *data, unsigned long type,
     }
 
     if (!s && type != OUT_RESERVE) {
-        error(ERR_WARNING, "attempt to initialise memory in the"
+        error(ERR_WARNING, "attempt to initialize memory in the"
               " BSS section: ignored");
         if (type == OUT_REL2ADR)
-            realbytes = 2;
+            size = 2;
         else if (type == OUT_REL4ADR)
-            realbytes = 4;
-        bsslen += realbytes;
+            size = 4;
+        bsslen += size;
         return;
     }
 
     if (type == OUT_RESERVE) {
         if (s) {
-            error(ERR_WARNING, "uninitialised space declared in"
+            error(ERR_WARNING, "uninitialized space declared in"
                   " %s section: zeroing",
                   (segto == stext.index ? "code" : "data"));
-            as86_sect_write(s, NULL, realbytes);
-            as86_add_piece(s, 0, 0L, 0L, realbytes, 0);
+            as86_sect_write(s, NULL, size);
+            as86_add_piece(s, 0, 0L, 0L, size, 0);
         } else
-            bsslen += realbytes;
+            bsslen += size;
     } else if (type == OUT_RAWDATA) {
         if (segment != NO_SEG)
             error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
-        as86_sect_write(s, data, realbytes);
-        as86_add_piece(s, 0, 0L, 0L, realbytes, 0);
+        as86_sect_write(s, data, size);
+        as86_add_piece(s, 0, 0L, 0L, size, 0);
     } else if (type == OUT_ADDRESS) {
         if (segment != NO_SEG) {
             if (segment % 2) {
                 error(ERR_NONFATAL, "as86 format does not support"
                       " segment base references");
             } else {
-                offset = *(long *)data;
-                as86_add_piece(s, 1, offset, segment, realbytes, 0);
+                offset = *(int64_t *)data;
+                as86_add_piece(s, 1, offset, segment, size, 0);
             }
         } else {
             p = mydata;
-            WRITELONG(p, *(long *)data);
-            as86_sect_write(s, data, realbytes);
-            as86_add_piece(s, 0, 0L, 0L, realbytes, 0);
+            WRITELONG(p, *(int64_t *)data);
+            as86_sect_write(s, data, size);
+            as86_add_piece(s, 0, 0L, 0L, size, 0);
         }
     } else if (type == OUT_REL2ADR) {
         if (segment == segto)
@@ -332,8 +337,8 @@ static void as86_out(long segto, const void *data, unsigned long type,
                 error(ERR_NONFATAL, "as86 format does not support"
                       " segment base references");
             } else {
-                offset = *(long *)data;
-                as86_add_piece(s, 1, offset - realbytes + 2, segment, 2L,
+                offset = *(int64_t *)data;
+                as86_add_piece(s, 1, offset - size + 2, segment, 2L,
                                1);
             }
         }
@@ -345,8 +350,8 @@ static void as86_out(long segto, const void *data, unsigned long type,
                 error(ERR_NONFATAL, "as86 format does not support"
                       " segment base references");
             } else {
-                offset = *(long *)data;
-                as86_add_piece(s, 1, offset - realbytes + 4, segment, 4L,
+                offset = *(int64_t *)data;
+                as86_add_piece(s, 1, offset - size + 4, segment, 4L,
                                1);
             }
         }
@@ -355,8 +360,8 @@ static void as86_out(long segto, const void *data, unsigned long type,
 
 static void as86_write(void)
 {
-    unsigned long i;
-    long symlen, seglen, segsize;
+    uint32_t i;
+    int32_t symlen, seglen, segsize;
 
     /*
      * First, go through the symbol records working out how big
@@ -385,11 +390,11 @@ static void as86_write(void)
      * descriptor word at the same time.
      */
     seglen = segsize = 0;
-    if ((unsigned long)stext.len > 65535L)
+    if ((uint32_t)stext.len > 65535L)
         segsize |= 0x03000000L, seglen += 4;
     else
         segsize |= 0x02000000L, seglen += 2;
-    if ((unsigned long)sdata.len > 65535L)
+    if ((uint32_t)sdata.len > 65535L)
         segsize |= 0xC0000000L, seglen += 4;
     else
         segsize |= 0x80000000L, seglen += 2;
@@ -397,23 +402,23 @@ static void as86_write(void)
     /*
      * Emit the as86 header.
      */
-    fwritelong(0x000186A3L, as86fp);
+    fwriteint32_t(0x000186A3L, as86fp);
     fputc(0x2A, as86fp);
-    fwritelong(27 + symlen + seglen + strslen, as86fp); /* header length */
-    fwritelong(stext.len + sdata.len + bsslen, as86fp);
-    fwriteshort(strslen, as86fp);
-    fwriteshort(0, as86fp);     /* class = revision = 0 */
-    fwritelong(0x55555555L, as86fp);    /* segment max sizes: always this */
-    fwritelong(segsize, as86fp);        /* segment size descriptors */
+    fwriteint32_t(27 + symlen + seglen + strslen, as86fp); /* header length */
+    fwriteint32_t(stext.len + sdata.len + bsslen, as86fp);
+    fwriteint16_t(strslen, as86fp);
+    fwriteint16_t(0, as86fp);     /* class = revision = 0 */
+    fwriteint32_t(0x55555555L, as86fp);    /* segment max sizes: always this */
+    fwriteint32_t(segsize, as86fp);        /* segment size descriptors */
     if (segsize & 0x01000000L)
-        fwritelong(stext.len, as86fp);
+        fwriteint32_t(stext.len, as86fp);
     else
-        fwriteshort(stext.len, as86fp);
+        fwriteint16_t(stext.len, as86fp);
     if (segsize & 0x40000000L)
-        fwritelong(sdata.len + bsslen, as86fp);
+        fwriteint32_t(sdata.len + bsslen, as86fp);
     else
-        fwriteshort(sdata.len + bsslen, as86fp);
-    fwriteshort(nsyms, as86fp);
+        fwriteint16_t(sdata.len + bsslen, as86fp);
+    fwriteint16_t(nsyms, as86fp);
 
     /*
      * Write the symbol table.
@@ -421,8 +426,8 @@ static void as86_write(void)
     saa_rewind(syms);
     for (i = 0; i < nsyms; i++) {
         struct Symbol *sym = saa_rstruct(syms);
-        fwriteshort(sym->strpos, as86fp);
-        fwriteshort(sym->flags, as86fp);
+        fwriteint16_t(sym->strpos, as86fp);
+        fwriteint16_t(sym->flags, as86fp);
         switch (sym->flags & (3 << 14)) {
         case 0 << 14:
             break;
@@ -430,10 +435,10 @@ static void as86_write(void)
             fputc(sym->value, as86fp);
             break;
         case 2 << 14:
-            fwriteshort(sym->value, as86fp);
+            fwriteint16_t(sym->value, as86fp);
             break;
         case 3 << 14:
-            fwritelong(sym->value, as86fp);
+            fwriteint32_t(sym->value, as86fp);
             break;
         }
     }
@@ -454,10 +459,10 @@ static void as86_write(void)
      */
     if (bsslen > 65535L) {
         fputc(0x13, as86fp);
-        fwritelong(bsslen, as86fp);
+        fwriteint32_t(bsslen, as86fp);
     } else if (bsslen > 255) {
         fputc(0x12, as86fp);
-        fwriteshort(bsslen, as86fp);
+        fwriteint16_t(bsslen, as86fp);
     } else if (bsslen) {
         fputc(0x11, as86fp);
         fputc(bsslen, as86fp);
@@ -488,8 +493,8 @@ static void as86_set_rsize(int size)
 static void as86_write_section(struct Section *sect, int index)
 {
     struct Piece *p;
-    unsigned long s;
-    long length;
+    uint32_t s;
+    int32_t length;
 
     fputc(0x20 + index, as86fp);        /* select the right section */
 
@@ -505,7 +510,7 @@ static void as86_write_section(struct Section *sect, int index)
             length = p->bytes;
             do {
                 char buf[64];
-                long tmplen = (length > 64 ? 64 : length);
+                int32_t tmplen = (length > 64 ? 64 : length);
                 fputc(0x40 | (tmplen & 0x3F), as86fp);
                 saa_rnbytes(sect->data, buf, tmplen);
                 fwrite(buf, 1, tmplen, as86fp);
@@ -521,9 +526,9 @@ static void as86_write_section(struct Section *sect, int index)
             as86_set_rsize(p->bytes);
             fputc(0x80 | (p->relative ? 0x20 : 0) | p->number, as86fp);
             if (as86_reloc_size == 2)
-                fwriteshort(p->offset, as86fp);
+                fwriteint16_t(p->offset, as86fp);
             else
-                fwritelong(p->offset, as86fp);
+                fwriteint32_t(p->offset, as86fp);
             break;
         case 2:
             /*
@@ -543,7 +548,7 @@ static void as86_write_section(struct Section *sect, int index)
                   (p->relative ? 0x20 : 0) |
                   (p->number > 255 ? 0x04 : 0) | s, as86fp);
             if (p->number > 255)
-                fwriteshort(p->number, as86fp);
+                fwriteint16_t(p->number, as86fp);
             else
                 fputc(p->number, as86fp);
             switch ((int)s) {
@@ -553,10 +558,10 @@ static void as86_write_section(struct Section *sect, int index)
                 fputc(p->offset, as86fp);
                 break;
             case 2:
-                fwriteshort(p->offset, as86fp);
+                fwriteint16_t(p->offset, as86fp);
                 break;
             case 3:
-                fwritelong(p->offset, as86fp);
+                fwriteint32_t(p->offset, as86fp);
                 break;
             }
             break;
@@ -564,19 +569,22 @@ static void as86_write_section(struct Section *sect, int index)
 }
 
 static void as86_sect_write(struct Section *sect,
-                            const unsigned char *data, unsigned long len)
+                            const uint8_t *data, uint32_t len)
 {
     saa_wbytes(sect->data, data, len);
     sect->datalen += len;
 }
 
-static long as86_segbase(long segment)
+static int32_t as86_segbase(int32_t segment)
 {
     return segment;
 }
 
 static int as86_directive(char *directive, char *value, int pass)
 {
+    (void)directive;
+    (void)value;
+    (void)pass;
     return 0;
 }
 
@@ -602,11 +610,18 @@ static const char *as86_stdmac[] = {
 
 static int as86_set_info(enum geninfo type, char **val)
 {
+    (void)type;
+    (void)val;
     return 0;
 }
-void as86_linenumber(char *name, long segment, long offset, int is_main,
+void as86_linenumber(char *name, int32_t segment, int32_t offset, int is_main,
                      int lineno)
 {
+    (void)name;
+    (void)segment;
+    (void)offset;
+    (void)is_main;
+    (void)lineno;
 }
 struct ofmt of_as86 = {
     "Linux as86 (bin86 version 0.3) object files",
