@@ -1,21 +1,22 @@
 /*
  * outrdf2.c	output routines for the Netwide Assembler to produce
- *		RDOFF version 2 format object files, which is used as a
- *		main binary format in the RadiOS (http://radios.sf.net).
- *		Originally Julian planned to use it in his MOSCOW
- *		operating system.
+ *		RDOFF version 2 format object files, which Julian originally
+ *		planned to use it in his MOSCOW operating system.
  *
  * The Netwide Assembler is copyright (C) 1996-1998 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  */
 
+#include "compiler.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
 #include <assert.h>
+#include <inttypes.h>
 
 #include "nasm.h"
 #include "nasmlib.h"
@@ -73,20 +74,25 @@ static efunc error;
 static struct seginfo {
     char *segname;
     int segnumber;
-    uint16 segtype;
-    uint16 segreserved;
-    long seglength;
+    uint16_t segtype;
+    uint16_t segreserved;
+    int32_t seglength;
 } segments[RDF_MAXSEGS];
 
 static int nsegments;
 
-static long bsslength;
-static long headerlength;
+static int32_t bsslength;
+static int32_t headerlength;
 
 static void rdf2_init(FILE * fp, efunc errfunc, ldfunc ldef, evalfunc eval)
 {
     int segtext, segdata, segbss;
 
+    (void)ldef;
+    (void)eval;
+
+    maxbits = 64;
+
     /* set up the initial segments */
     segments[0].segname = ".text";
     segments[0].segnumber = 0;
@@ -128,13 +134,16 @@ static void rdf2_init(FILE * fp, efunc errfunc, ldfunc ldef, evalfunc eval)
     headerlength = 0;
 }
 
-static long rdf2_section_names(char *name, int pass, int *bits)
+static int32_t rdf2_section_names(char *name, int pass, int *bits)
 {
     int i;
+    bool err;
     char *p, *q;
     int code = -1;
     int reserved = 0;
 
+    (void)pass;
+
     /*
      * Default is 32 bits, in the text segment.
      */
@@ -159,15 +168,15 @@ static long rdf2_section_names(char *name, int pass, int *bits)
         if ((q = strchr(p, ','))) {
             *q++ = '\0';
 
-            reserved = readnum(q, &i);
-            if (i) {
+            reserved = readnum(q, &err);
+            if (err) {
                 error(ERR_NONFATAL,
                       "value following comma must be numeric");
                 reserved = 0;
             }
         }
         /*
-         * check it against the text strings in segmenttypes 
+         * check it against the text strings in segmenttypes
          */
 
         for (i = 0; i < COUNT_SEGTYPES; i++)
@@ -176,8 +185,8 @@ static long rdf2_section_names(char *name, int pass, int *bits)
                 break;
             }
         if (code == -1) {       /* didn't find anything */
-            code = readnum(p, &i);
-            if (i) {
+            code = readnum(p, &err);
+            if (err) {
                 error(ERR_NONFATAL, "unrecognised RDF segment type (%s)",
                       p);
                 code = 3;
@@ -226,7 +235,7 @@ static void write_reloc_rec(struct RelocRec *r)
 {
     char buf[4], *b;
 
-    if (r->refseg != (uint16) NO_SEG && (r->refseg & 1))        /* segment base ref */
+    if (r->refseg != (uint16_t) NO_SEG && (r->refseg & 1))        /* segment base ref */
         r->type = RDFREC_SEGRELOC;
 
     r->refseg >>= 1;            /* adjust segment nos to RDF rather than NASM */
@@ -344,7 +353,7 @@ static void write_modname_rec(struct ModRec *r)
 /*
  * Handle export, import and common records.
  */
-static void rdf2_deflabel(char *name, long segment, long offset,
+static void rdf2_deflabel(char *name, int32_t segment, int64_t offset,
                           int is_global, char *special)
 {
     struct ExportRec r;
@@ -352,7 +361,7 @@ static void rdf2_deflabel(char *name, long segment, long offset,
     struct CommonRec ci;
     static int farsym = 0;
     static int i;
-    byte symflags = 0;
+    char symflags = 0;
     int len;
 
     /* Check if the label length is OK */
@@ -379,7 +388,7 @@ static void rdf2_deflabel(char *name, long segment, long offset,
          * of two; if so, store it as the alignment for the common variable.
          */
         if (special) {
-            int err;
+            bool err;
             ci.align = readnum(special, &err);
             if (err)
                 error(ERR_NONFATAL, "alignment constraint `%s' is not a"
@@ -475,9 +484,9 @@ static void membufwrite(int segment, const void *data, int bytes)
     if (bytes < 0) {
         b = buf;
         if (bytes == -2)
-            WRITESHORT(b, *(short *)data);
+            WRITESHORT(b, *(int16_t *)data);
         else
-            WRITELONG(b, *(long *)data);
+            WRITELONG(b, *(int32_t *)data);
         data = buf;
         bytes = -bytes;
     }
@@ -498,16 +507,16 @@ static int getsegmentlength(int segment)
     return segments[i].seglength;
 }
 
-static void rdf2_out(long segto, const void *data, unsigned long type,
-                     long segment, long wrt)
+static void rdf2_out(int32_t segto, const void *data,
+		     enum out_type type, uint64_t size,
+                     int32_t segment, int32_t wrt)
 {
-    long bytes = type & OUT_SIZMASK;
     struct RelocRec rr;
-    unsigned char databuf[4], *pd;
+    uint8_t databuf[8], *pd;
     int seg;
 
     if (segto == NO_SEG) {
-        if ((type & OUT_TYPMASK) != OUT_RESERVE)
+        if (type != OUT_RESERVE)
             error(ERR_NONFATAL,
                   "attempt to assemble code in ABSOLUTE space");
         return;
@@ -530,31 +539,29 @@ static void rdf2_out(long segto, const void *data, unsigned long type,
         error(ERR_NONFATAL, "WRT not supported by rdf output format");
     }
 
-    type &= OUT_TYPMASK;
-
     if (segto == 2 && type != OUT_RESERVE) {
-        error(ERR_NONFATAL, "BSS segments may not be initialised");
+        error(ERR_NONFATAL, "BSS segments may not be initialized");
 
         /* just reserve the space for now... */
 
         if (type == OUT_REL2ADR)
-            bytes = 2;
+            size = 2;
         else
-            bytes = 4;
+            size = 4;
         type = OUT_RESERVE;
     }
 
     if (type == OUT_RESERVE) {
         if (segto == 2)         /* BSS segment space reserverd */
-            bsslength += bytes;
+            bsslength += size;
         else
-            while (bytes--)
+            while (size--)
                 membufwrite(segto, databuf, 1);
     } else if (type == OUT_RAWDATA) {
         if (segment != NO_SEG)
             error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
 
-        membufwrite(segto, data, bytes);
+        membufwrite(segto, data, size);
     } else if (type == OUT_ADDRESS) {
 
         /* if segment == NO_SEG then we are writing an address of an
@@ -570,19 +577,14 @@ static void rdf2_out(long segto, const void *data, unsigned long type,
             rr.reclen = 8;
             rr.segment = segto; /* segment we're currently in */
             rr.offset = getsegmentlength(segto);        /* current offset */
-            rr.length = bytes;  /* length of reference */
+            rr.length = size;  /* length of reference */
             rr.refseg = segment;        /* segment referred to */
             write_reloc_rec(&rr);
         }
 
         pd = databuf;           /* convert address to little-endian */
-        if (bytes == 2)
-            WRITESHORT(pd, *(long *)data);
-        else
-            WRITELONG(pd, *(long *)data);
-
-        membufwrite(segto, databuf, bytes);
-
+	WRITEADDR(pd, *(int64_t *)data, size);
+        membufwrite(segto, databuf, size);
     } else if (type == OUT_REL2ADR) {
         if (segment == segto)
             error(ERR_PANIC, "intra-segment OUT_REL2ADR");
@@ -600,7 +602,7 @@ static void rdf2_out(long segto, const void *data, unsigned long type,
             /* what do we put in the code? Simply the data. This should almost
              * always be zero, unless someone's doing segment arithmetic...
              */
-            rr.offset = *(long *)data;
+            rr.offset = *(int64_t *)data;
         } else {
             rr.type = RDFREC_RELOC;     /* type signature */
             rr.segment = segto + 64;    /* segment we're currently in + rel flag */
@@ -611,12 +613,12 @@ static void rdf2_out(long segto, const void *data, unsigned long type,
              * address of imported symbol onto it to get address relative to end of
              * instruction: import_address + data(offset) - end_of_instrn */
 
-            rr.offset = *(long *)data - (rr.offset + bytes);
+            rr.offset = *(int32_t *)data - (rr.offset + size);
         }
 
         membufwrite(segto, &rr.offset, -2);
     } else if (type == OUT_REL4ADR) {
-        if (segment == segto)
+        if ((segment == segto) && (globalbits != 64))
             error(ERR_PANIC, "intra-segment OUT_REL4ADR");
         if (segment != NO_SEG && segment % 2) {
             error(ERR_PANIC, "erm... 4 byte segment base ref?");
@@ -630,7 +632,7 @@ static void rdf2_out(long segto, const void *data, unsigned long type,
         rr.reclen = 8;
         write_reloc_rec(&rr);
 
-        rr.offset = *(long *)data - (rr.offset + bytes);
+        rr.offset = *(int64_t *)data - (rr.offset + size);
 
         membufwrite(segto, &rr.offset, -4);
     }
@@ -638,7 +640,7 @@ static void rdf2_out(long segto, const void *data, unsigned long type,
 
 static void rdf2_cleanup(int debuginfo)
 {
-    long l;
+    int32_t l;
     struct BSSRec bs;
     int i;
 
@@ -668,9 +670,9 @@ static void rdf2_cleanup(int debuginfo)
     }
     l += 10;                    /* null segment */
 
-    fwritelong(l, ofile);
+    fwriteint32_t(l, ofile);
 
-    fwritelong(headerlength, ofile);
+    fwriteint32_t(headerlength, ofile);
     saa_fpwrite(header, ofile); /* dump header */
     saa_free(header);
 
@@ -678,24 +680,24 @@ static void rdf2_cleanup(int debuginfo)
         if (i == 2)
             continue;
 
-        fwriteshort(segments[i].segtype, ofile);
-        fwriteshort(segments[i].segnumber, ofile);
-        fwriteshort(segments[i].segreserved, ofile);
-        fwritelong(segments[i].seglength, ofile);
+        fwriteint16_t(segments[i].segtype, ofile);
+        fwriteint16_t(segments[i].segnumber, ofile);
+        fwriteint16_t(segments[i].segreserved, ofile);
+        fwriteint32_t(segments[i].seglength, ofile);
 
         saa_fpwrite(seg[i], ofile);
         saa_free(seg[i]);
     }
 
     /* null segment - write 10 bytes of zero */
-    fwritelong(0, ofile);
-    fwritelong(0, ofile);
-    fwriteshort(0, ofile);
+    fwriteint32_t(0, ofile);
+    fwriteint32_t(0, ofile);
+    fwriteint16_t(0, ofile);
 
     fclose(ofile);
 }
 
-static long rdf2_segbase(long segment)
+static int32_t rdf2_segbase(int32_t segment)
 {
     return segment;
 }
@@ -758,6 +760,8 @@ static const char *rdf2_stdmac[] = {
 
 static int rdf2_set_info(enum geninfo type, char **val)
 {
+    (void)type;
+    (void)val;
     return 0;
 }
 
