@@ -2,34 +2,32 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  *
  * initial version 27/iii/95 by Simon Tatham
  */
 
+#include "compiler.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <stddef.h>
 #include <string.h>
 #include <ctype.h>
+#include <inttypes.h>
 
 #include "nasm.h"
 #include "insns.h"
 #include "nasmlib.h"
+#include "stdscan.h"
 #include "parser.h"
 #include "float.h"
+#include "tables.h"
 
 extern int in_abs_seg;          /* ABSOLUTE segment flag */
-extern long abs_seg;            /* ABSOLUTE segment */
-extern long abs_offset;         /* ABSOLUTE segment offset */
-
-#include "regflags.c"           /* List of register flags */
-
-enum {                          /* special tokens */
-    S_BYTE, S_DWORD, S_FAR, S_LONG, S_NEAR, S_NOSPLIT, S_QWORD,
-    S_SHORT, S_STRICT, S_TO, S_TWORD, S_WORD
-};
+extern int32_t abs_seg;            /* ABSOLUTE segment */
+extern int32_t abs_offset;         /* ABSOLUTE segment offset */
 
 static int is_comma_next(void);
 
@@ -37,22 +35,147 @@ static int i;
 static struct tokenval tokval;
 static efunc error;
 static struct ofmt *outfmt;     /* Structure of addresses of output routines */
-static loc_t *location;         /* Pointer to current line's segment,offset */
+static struct location *location;         /* Pointer to current line's segment,offset */
 
-void parser_global_info(struct ofmt *output, loc_t * locp)
+void parser_global_info(struct ofmt *output, struct location * locp)
 {
     outfmt = output;
     location = locp;
 }
 
+static int prefix_slot(enum prefixes prefix)
+{
+    switch (prefix) {
+    case R_CS:
+    case R_DS:
+    case R_SS:
+    case R_ES:
+    case R_FS:
+    case R_GS:
+	return PPS_SEG;
+    case P_LOCK:
+    case P_REP:
+    case P_REPE:
+    case P_REPZ:
+    case P_REPNE:
+    case P_REPNZ:
+	return PPS_LREP;
+    case P_O16:
+    case P_O32:
+    case P_O64:
+    case P_OSP:
+	return PPS_OSIZE;
+    case P_A16:
+    case P_A32:
+    case P_A64:
+    case P_ASP:
+	return PPS_ASIZE;
+    default:
+	error(ERR_PANIC, "Invalid value %d passed to prefix_slot()", prefix);
+	return -1;
+    }
+}
+
+static void process_size_override(insn * result, int operand)
+{
+    if (tasm_compatible_mode) {
+	switch ((int)tokval.t_integer) {
+	    /* For TASM compatibility a size override inside the
+	     * brackets changes the size of the operand, not the
+	     * address type of the operand as it does in standard
+	     * NASM syntax. Hence:
+	     *
+	     *  mov     eax,[DWORD val]
+	     *
+	     * is valid syntax in TASM compatibility mode. Note that
+	     * you lose the ability to override the default address
+	     * type for the instruction, but we never use anything
+	     * but 32-bit flat model addressing in our code.
+	     */
+	case S_BYTE:
+	    result->oprs[operand].type |= BITS8;
+	    break;
+	case S_WORD:
+	    result->oprs[operand].type |= BITS16;
+	    break;
+	case S_DWORD:
+	case S_LONG:
+	    result->oprs[operand].type |= BITS32;
+	    break;
+	case S_QWORD:
+	    result->oprs[operand].type |= BITS64;
+	    break;
+	case S_TWORD:
+	    result->oprs[operand].type |= BITS80;
+	    break;
+	case S_OWORD:
+	    result->oprs[operand].type |= BITS128;
+	    break;
+	default:
+	    error(ERR_NONFATAL,
+		  "invalid operand size specification");
+	    break;
+	}
+    } else {
+	/* Standard NASM compatible syntax */
+	switch ((int)tokval.t_integer) {
+	case S_NOSPLIT:
+	    result->oprs[operand].eaflags |= EAF_TIMESTWO;
+	    break;
+	case S_REL:
+	    result->oprs[operand].eaflags |= EAF_REL;
+	    break;
+	case S_ABS:
+	    result->oprs[operand].eaflags |= EAF_ABS;
+	    break;
+	case S_BYTE:
+	    result->oprs[operand].disp_size = 8;
+	    result->oprs[operand].eaflags |= EAF_BYTEOFFS;
+	    break;
+	case P_A16:
+	case P_A32:
+	case P_A64:
+	    if (result->prefixes[PPS_ASIZE] &&
+		result->prefixes[PPS_ASIZE] != tokval.t_integer)
+		error(ERR_NONFATAL,
+		      "conflicting address size specifications");
+	    else
+		result->prefixes[PPS_ASIZE] = tokval.t_integer;
+	    break;
+	case S_WORD:
+	    result->oprs[operand].disp_size = 16;
+	    result->oprs[operand].eaflags |= EAF_WORDOFFS;
+	    break;
+	case S_DWORD:
+	case S_LONG:
+	    result->oprs[operand].disp_size = 32;
+	    result->oprs[operand].eaflags |= EAF_WORDOFFS;
+	    break;
+	case S_QWORD:
+	    result->oprs[operand].disp_size = 64;
+	    result->oprs[operand].eaflags |= EAF_WORDOFFS;
+	    break;
+	default:
+	    error(ERR_NONFATAL, "invalid size specification in"
+		  " effective address");
+	    break;
+	}
+    }
+}
+
 insn *parse_line(int pass, char *buffer, insn * result,
                  efunc errfunc, evalfunc evaluate, ldfunc ldef)
 {
     int operand;
     int critical;
     struct eval_hints hints;
+    int j;
+    bool first;
+    bool insn_is_label = false;
 
-    result->forw_ref = FALSE;
+restart_parse:
+    first = true;
+    result->forw_ref = false;
     error = errfunc;
 
     stdscan_reset();
@@ -61,21 +184,23 @@ insn *parse_line(int pass, char *buffer, insn * result,
 
     result->label = NULL;       /* Assume no label */
     result->eops = NULL;        /* must do this, whatever happens */
-    result->operands = 0;       /* must initialise this */
+    result->operands = 0;       /* must initialize this */
 
     if (i == 0) {               /* blank line - ignore */
         result->opcode = -1;    /* and no instruction either */
         return result;
     }
     if (i != TOKEN_ID && i != TOKEN_INSN && i != TOKEN_PREFIX &&
-        (i != TOKEN_REG || (REG_SREG & ~reg_flags[tokval.t_integer]))) {
+        (i != TOKEN_REG || (REG_SREG & ~nasm_reg_flags[tokval.t_integer]))) {
         error(ERR_NONFATAL, "label or instruction expected"
               " at start of line");
         result->opcode = -1;
         return result;
     }
 
-    if (i == TOKEN_ID) {        /* there's a label here */
+    if (i == TOKEN_ID || (insn_is_label && i == TOKEN_INSN)) {
+        /* there's a label here */
+	first = false;
         result->label = tokval.t_charptr;
         i = stdscan(NULL, &tokval);
         if (i == ':') {         /* skip over the optional colon */
@@ -93,7 +218,7 @@ insn *parse_line(int pass, char *buffer, insn * result,
              * am still not certain.
              */
             ldef(result->label, in_abs_seg ? abs_seg : location->segment,
-                 location->offset, NULL, TRUE, FALSE, outfmt, errfunc);
+                 location->offset, NULL, true, false, outfmt, errfunc);
         }
     }
 
@@ -102,12 +227,15 @@ insn *parse_line(int pass, char *buffer, insn * result,
         return result;
     }
 
-    result->nprefix = 0;
+    for (j = 0; j < MAXPREFIX; j++)
+	result->prefixes[j] = P_none;
     result->times = 1L;
 
     while (i == TOKEN_PREFIX ||
-           (i == TOKEN_REG && !(REG_SREG & ~reg_flags[tokval.t_integer])))
+           (i == TOKEN_REG && !(REG_SREG & ~nasm_reg_flags[tokval.t_integer])))
     {
+	first = false;
+
         /*
          * Handle special case: the TIMES prefix.
          */
@@ -135,17 +263,29 @@ insn *parse_line(int pass, char *buffer, insn * result,
                 }
             }
         } else {
-            if (result->nprefix == MAXPREFIX)
-                error(ERR_NONFATAL,
-                      "instruction has more than %d prefixes", MAXPREFIX);
-            else
-                result->prefixes[result->nprefix++] = tokval.t_integer;
+	    int slot = prefix_slot(tokval.t_integer);
+	    if (result->prefixes[slot]) {
+               if (result->prefixes[slot] == tokval.t_integer)
+		    error(ERR_WARNING,
+		      "instruction has redundant prefixes");
+               else
+		    error(ERR_NONFATAL,
+		      "instruction has conflicting prefixes");
+	    }
+	    result->prefixes[slot] = tokval.t_integer;
             i = stdscan(NULL, &tokval);
         }
     }
 
     if (i != TOKEN_INSN) {
-        if (result->nprefix > 0 && i == 0) {
+	int j;
+	enum prefixes pfx;
+
+	for (j = 0; j < MAXPREFIX; j++)
+	    if ((pfx = result->prefixes[j]) != P_none)
+		break;
+
+        if (i == 0 && pfx != P_none) {
             /*
              * Instruction prefixes are present, but no actual
              * instruction. This is allowed: at this point we
@@ -178,28 +318,37 @@ insn *parse_line(int pass, char *buffer, insn * result,
      * For the moment, EQU has the same difficulty, so we'll
      * include that.
      */
-    if (result->opcode == I_RESB || result->opcode == I_RESW || result->opcode == I_RESD || result->opcode == I_RESQ || result->opcode == I_REST || result->opcode == I_EQU || result->opcode == I_INCBIN) {    /* fbk */
-        critical = pass0;
+    if (result->opcode == I_RESB || result->opcode == I_RESW ||
+	result->opcode == I_RESD || result->opcode == I_RESQ ||
+	result->opcode == I_REST || result->opcode == I_RESO ||
+	result->opcode == I_RESY ||
+	result->opcode == I_INCBIN) {
+        critical = (pass0 < 2 ? 1 : 2);
+
     } else
         critical = (pass == 2 ? 2 : 0);
 
-    if (result->opcode == I_DB ||
-        result->opcode == I_DW ||
-        result->opcode == I_DD ||
-        result->opcode == I_DQ ||
-        result->opcode == I_DT || result->opcode == I_INCBIN) {
+    if (result->opcode == I_DB || result->opcode == I_DW ||
+        result->opcode == I_DD || result->opcode == I_DQ ||
+        result->opcode == I_DT || result->opcode == I_DO ||
+	result->opcode == I_DY || result->opcode == I_INCBIN) {
         extop *eop, **tail = &result->eops, **fixptr;
         int oper_num = 0;
 
-        result->eops_float = FALSE;
+        result->eops_float = false;
 
         /*
-         * Begin to read the DB/DW/DD/DQ/DT/INCBIN operands.
+         * Begin to read the DB/DW/DD/DQ/DT/DO/INCBIN operands.
          */
         while (1) {
             i = stdscan(NULL, &tokval);
             if (i == 0)
                 break;
+	    else if (first && i == ':') {
+		insn_is_label = true;
+		goto restart_parse;
+	    }
+	    first = false;
             fixptr = tail;
             eop = *tail = nasm_malloc(sizeof(extop));
             tail = &eop->next;
@@ -215,47 +364,66 @@ insn *parse_line(int pass, char *buffer, insn * result,
                 continue;
             }
 
-            if ((i == TOKEN_FLOAT && is_comma_next()) || i == '-') {
-                long sign = +1L;
+            if ((i == TOKEN_FLOAT && is_comma_next())
+		|| i == '-' || i == '+') {
+                int32_t sign = +1;
 
-                if (i == '-') {
+                if (i == '+' || i == '-') {
                     char *save = stdscan_bufptr;
+		    int token = i;
+		    sign = (i == '-') ? -1 : 1;
                     i = stdscan(NULL, &tokval);
-                    sign = -1L;
                     if (i != TOKEN_FLOAT || !is_comma_next()) {
                         stdscan_bufptr = save;
-                        i = tokval.t_type = '-';
+                        i = tokval.t_type = token;
                     }
                 }
 
                 if (i == TOKEN_FLOAT) {
                     eop->type = EOT_DB_STRING;
-                    result->eops_float = TRUE;
-                    if (result->opcode == I_DD)
+                    result->eops_float = true;
+		    switch (result->opcode) {
+		    case I_DB:
+			eop->stringlen = 1;
+			break;
+		    case I_DW:
+			eop->stringlen = 2;
+			break;
+		    case I_DD:
                         eop->stringlen = 4;
-                    else if (result->opcode == I_DQ)
+			break;
+		    case I_DQ:
                         eop->stringlen = 8;
-                    else if (result->opcode == I_DT)
+			break;
+		    case I_DT:
                         eop->stringlen = 10;
-                    else {
+			break;
+		    case I_DO:
+                        eop->stringlen = 16;
+			break;
+		    case I_DY:
                         error(ERR_NONFATAL, "floating-point constant"
-                              " encountered in `D%c' instruction",
-                              result->opcode == I_DW ? 'W' : 'B');
+                              " encountered in DY instruction");
+			eop->stringlen = 0;
+			break;
+		    default:
+                        error(ERR_NONFATAL, "floating-point constant"
+                              " encountered in unknown instruction");
                         /*
                          * fix suggested by Pedro Gimeno... original line
                          * was:
                          * eop->type = EOT_NOTHING;
                          */
                         eop->stringlen = 0;
+			break;
                     }
-                    eop =
-                        nasm_realloc(eop, sizeof(extop) + eop->stringlen);
+                    eop = nasm_realloc(eop, sizeof(extop) + eop->stringlen);
                     tail = &eop->next;
                     *fixptr = eop;
                     eop->stringval = (char *)eop + sizeof(extop);
-                    if (eop->stringlen < 4 ||
+                    if (!eop->stringlen ||
                         !float_const(tokval.t_charptr, sign,
-                                     (unsigned char *)eop->stringval,
+                                     (uint8_t *)eop->stringval,
                                      eop->stringlen, error))
                         eop->type = EOT_NOTHING;
                     i = stdscan(NULL, &tokval); /* eat the comma */
@@ -342,22 +510,27 @@ insn *parse_line(int pass, char *buffer, insn * result,
         return result;
     }
 
-    /* right. Now we begin to parse the operands. There may be up to three
+    /* right. Now we begin to parse the operands. There may be up to four
      * of these, separated by commas, and terminated by a zero token. */
 
-    for (operand = 0; operand < 3; operand++) {
-        expr *value;            /* used most of the time */
+    for (operand = 0; operand < MAX_OPERANDS; operand++) {
+	expr *value;		/* used most of the time */
         int mref;               /* is this going to be a memory ref? */
         int bracket;            /* is it a [] mref, or a & mref? */
         int setsize = 0;
 
-        result->oprs[operand].addr_size = 0;    /* have to zero this whatever */
+        result->oprs[operand].disp_size = 0;    /* have to zero this whatever */
         result->oprs[operand].eaflags = 0;      /* and this */
         result->oprs[operand].opflags = 0;
 
         i = stdscan(NULL, &tokval);
         if (i == 0)
             break;              /* end of operands: get out of here */
+	else if (first && i == ':') {
+	    insn_is_label = true;
+	    goto restart_parse;
+	}
+	first = false;
         result->oprs[operand].type = 0; /* so far, no override */
         while (i == TOKEN_SPECIAL) {    /* size specifiers */
             switch ((int)tokval.t_integer) {
@@ -387,6 +560,16 @@ insn *parse_line(int pass, char *buffer, insn * result,
                     result->oprs[operand].type |= BITS80;
                 setsize = 1;
                 break;
+            case S_OWORD:
+                if (!setsize)
+                    result->oprs[operand].type |= BITS128;
+                setsize = 1;
+                break;
+            case S_YWORD:
+                if (!setsize)
+                    result->oprs[operand].type |= BITS256;
+                setsize = 1;
+                break;
             case S_TO:
                 result->oprs[operand].type |= TO;
                 break;
@@ -409,72 +592,16 @@ insn *parse_line(int pass, char *buffer, insn * result,
         }
 
         if (i == '[' || i == '&') {     /* memory reference */
-            mref = TRUE;
+            mref = true;
             bracket = (i == '[');
-            i = stdscan(NULL, &tokval);
-            if (i == TOKEN_SPECIAL) {   /* check for address size override */
-                if (tasm_compatible_mode) {
-                    switch ((int)tokval.t_integer) {
-                        /* For TASM compatibility a size override inside the
-                         * brackets changes the size of the operand, not the
-                         * address type of the operand as it does in standard
-                         * NASM syntax. Hence:
-                         *
-                         *  mov     eax,[DWORD val]
-                         *
-                         * is valid syntax in TASM compatibility mode. Note that
-                         * you lose the ability to override the default address
-                         * type for the instruction, but we never use anything
-                         * but 32-bit flat model addressing in our code.
-                         */
-                    case S_BYTE:
-                        result->oprs[operand].type |= BITS8;
-                        break;
-                    case S_WORD:
-                        result->oprs[operand].type |= BITS16;
-                        break;
-                    case S_DWORD:
-                    case S_LONG:
-                        result->oprs[operand].type |= BITS32;
-                        break;
-                    case S_QWORD:
-                        result->oprs[operand].type |= BITS64;
-                        break;
-                    case S_TWORD:
-                        result->oprs[operand].type |= BITS80;
-                        break;
-                    default:
-                        error(ERR_NONFATAL,
-                              "invalid operand size specification");
-                    }
-                } else {
-                    /* Standard NASM compatible syntax */
-                    switch ((int)tokval.t_integer) {
-                    case S_NOSPLIT:
-                        result->oprs[operand].eaflags |= EAF_TIMESTWO;
-                        break;
-                    case S_BYTE:
-                        result->oprs[operand].eaflags |= EAF_BYTEOFFS;
-                        break;
-                    case S_WORD:
-                        result->oprs[operand].addr_size = 16;
-                        result->oprs[operand].eaflags |= EAF_WORDOFFS;
-                        break;
-                    case S_DWORD:
-                    case S_LONG:
-                        result->oprs[operand].addr_size = 32;
-                        result->oprs[operand].eaflags |= EAF_WORDOFFS;
-                        break;
-                    default:
-                        error(ERR_NONFATAL, "invalid size specification in"
-                              " effective address");
-                    }
-                }
+            i = stdscan(NULL, &tokval); /* then skip the colon */
+            while (i == TOKEN_SPECIAL || i == TOKEN_PREFIX) {
+		process_size_override(result, operand);
                 i = stdscan(NULL, &tokval);
             }
         } else {                /* immediate operand, or register */
-            mref = FALSE;
-            bracket = FALSE;    /* placate optimisers */
+            mref = false;
+            bracket = false;    /* placate optimisers */
         }
 
         if ((result->oprs[operand].type & FAR) && !mref &&
@@ -487,7 +614,7 @@ insn *parse_line(int pass, char *buffer, insn * result,
                          critical, error, &hints);
         i = tokval.t_type;
         if (result->oprs[operand].opflags & OPFLAG_FORWARD) {
-            result->forw_ref = TRUE;
+            result->forw_ref = true;
         }
         if (!value) {           /* error in evaluator */
             result->opcode = -1;        /* unrecoverable parse error: */
@@ -498,28 +625,20 @@ insn *parse_line(int pass, char *buffer, insn * result,
              * Process the segment override.
              */
             if (value[1].type != 0 || value->value != 1 ||
-                REG_SREG & ~reg_flags[value->type])
+                REG_SREG & ~nasm_reg_flags[value->type])
                 error(ERR_NONFATAL, "invalid segment override");
-            else if (result->nprefix == MAXPREFIX)
+            else if (result->prefixes[PPS_SEG])
                 error(ERR_NONFATAL,
-                      "instruction has more than %d prefixes", MAXPREFIX);
-            else
-                result->prefixes[result->nprefix++] = value->type;
+                      "instruction has conflicting segment overrides");
+            else {
+		result->prefixes[PPS_SEG] = value->type;
+		if (!(REG_FSGS & ~nasm_reg_flags[value->type]))
+		    result->oprs[operand].eaflags |= EAF_FSGS;
+	    }
 
             i = stdscan(NULL, &tokval); /* then skip the colon */
-            if (i == TOKEN_SPECIAL) {   /* another check for size override */
-                switch ((int)tokval.t_integer) {
-                case S_WORD:
-                    result->oprs[operand].addr_size = 16;
-                    break;
-                case S_DWORD:
-                case S_LONG:
-                    result->oprs[operand].addr_size = 32;
-                    break;
-                default:
-                    error(ERR_NONFATAL, "invalid size specification in"
-                          " effective address");
-                }
+            while (i == TOKEN_SPECIAL || i == TOKEN_PREFIX) {
+		process_size_override(result, operand);
                 i = stdscan(NULL, &tokval);
             }
             value = evaluate(stdscan, NULL, &tokval,
@@ -527,7 +646,7 @@ insn *parse_line(int pass, char *buffer, insn * result,
                              critical, error, &hints);
             i = tokval.t_type;
             if (result->oprs[operand].opflags & OPFLAG_FORWARD) {
-                result->forw_ref = TRUE;
+                result->forw_ref = true;
             }
             /* and get the offset */
             if (!value) {       /* but, error in evaluator */
@@ -560,7 +679,7 @@ insn *parse_line(int pass, char *buffer, insn * result,
         if (mref) {             /* it's a memory reference */
             expr *e = value;
             int b, i, s;        /* basereg, indexreg, scale */
-            long o;             /* offset */
+            int64_t o;             /* offset */
 
             b = i = -1, o = s = 0;
             result->oprs[operand].hintbase = hints.base;
@@ -655,9 +774,18 @@ insn *parse_line(int pass, char *buffer, insn * result,
                 return result;
             }
 
-            result->oprs[operand].type |= MEMORY;
-            if (b == -1 && (i == -1 || s == 0))
-                result->oprs[operand].type |= MEM_OFFS;
+	    /* It is memory, but it can match any r/m operand */
+            result->oprs[operand].type |= MEMORY_ANY;
+
+	    if (b == -1 && (i == -1 || s == 0)) {
+		int is_rel = globalbits == 64 &&
+		    !(result->oprs[operand].eaflags & EAF_ABS) &&
+		    ((globalrel &&
+		      !(result->oprs[operand].eaflags & EAF_FSGS)) ||
+		     (result->oprs[operand].eaflags & EAF_REL));
+
+		result->oprs[operand].type |= is_rel ? IP_REL : MEM_OFFS;
+	    }
             result->oprs[operand].basereg = b;
             result->oprs[operand].indexreg = i;
             result->oprs[operand].scale = s;
@@ -679,12 +807,20 @@ insn *parse_line(int pass, char *buffer, insn * result,
                         result->oprs[operand].type |= UNITY;
                     if (optimizing >= 0 &&
                         !(result->oprs[operand].type & STRICT)) {
-                        if (reloc_value(value) >= -128 &&
-                            reloc_value(value) <= 127)
-                            result->oprs[operand].type |= SBYTE;
+			int64_t v64 = reloc_value(value);
+			int32_t v32 = (int32_t)v64;
+			int16_t v16 = (int16_t)v32;
+
+			if (v64 >= -128 && v64 <= 127)
+                            result->oprs[operand].type |= SBYTE64;
+			if (v32 >= -128 && v32 <= 127)
+                            result->oprs[operand].type |= SBYTE32;
+			if (v16 >= -128 && v16 <= 127)
+                            result->oprs[operand].type |= SBYTE16;
                     }
                 }
             } else {            /* it's a register */
+		unsigned int rs;
 
                 if (value->type >= EXPR_SIMPLE || value->value != 1) {
                     error(ERR_NONFATAL, "invalid operand type");
@@ -708,16 +844,16 @@ insn *parse_line(int pass, char *buffer, insn * result,
                      * we want to produce a warning iff the specified size
                      * is different from the register size
                      */
-                    i = result->oprs[operand].type & SIZE_MASK;
+                    rs = result->oprs[operand].type & SIZE_MASK;
                 } else
-                    i = 0;
+                    rs = 0;
 
                 result->oprs[operand].type &= TO;
                 result->oprs[operand].type |= REGISTER;
-                result->oprs[operand].type |= reg_flags[value->type];
+                result->oprs[operand].type |= nasm_reg_flags[value->type];
                 result->oprs[operand].basereg = value->type;
 
-                if (i && (result->oprs[operand].type & SIZE_MASK) != i)
+                if (rs && (result->oprs[operand].type & SIZE_MASK) != rs)
                     error(ERR_WARNING | ERR_PASS1,
                           "register size specification ignored");
             }
@@ -726,11 +862,12 @@ insn *parse_line(int pass, char *buffer, insn * result,
 
     result->operands = operand; /* set operand count */
 
-    while (operand < 3)         /* clear remaining operands */
-        result->oprs[operand++].type = 0;
+/* clear remaining operands */
+while (operand < MAX_OPERANDS)
+    result->oprs[operand++].type = 0;
 
     /*
-     * Transform RESW, RESD, RESQ, REST into RESB.
+     * Transform RESW, RESD, RESQ, REST, RESO, RESY into RESB.
      */
     switch (result->opcode) {
     case I_RESW:
@@ -749,6 +886,16 @@ insn *parse_line(int pass, char *buffer, insn * result,
         result->opcode = I_RESB;
         result->oprs[0].offset *= 10;
         break;
+    case I_RESO:
+        result->opcode = I_RESB;
+        result->oprs[0].offset *= 16;
+        break;
+    case I_RESY:
+        result->opcode = I_RESB;
+        result->oprs[0].offset *= 32;
+        break;
+    default:
+	break;
     }
 
     return result;
