@@ -7,7 +7,7 @@
  *
  * When a library is being created, special signature block is placed
  * in the beginning of the file. It is a string 'RDLIB' followed by a
- * version number, then long content size and a long time stamp.
+ * version number, then int32_t content size and a int32_t time stamp.
  * The module name of the signature block is '.sig'.
  *
  *
@@ -15,20 +15,23 @@
  * The format of the directory will be 'RDLDD' followed by a version
  * number, followed by the length of the directory, and then the
  * directory, the format of which has not yet been designed.
- * The module name of the directory must be '.dir'. 
+ * The module name of the directory must be '.dir'.
  *
  * All module names beginning with '.' are reserved for possible future
  * extensions. The linker ignores all such modules, assuming they have
- * the format of a six byte type & version identifier followed by long
+ * the format of a six uint8_t type & version identifier followed by int32_t
  * content size, followed by data.
  */
 
+#include "compiler.h"
+
 #include <stdio.h>
 #include <stdlib.h>
-#include <unistd.h>
 #include <errno.h>
 #include <string.h>
+#include <inttypes.h>
 #include <time.h>
+#include <inttypes.h>
 
 /* functions supported:
  *   create a library	(no extra operands required)
@@ -56,11 +59,11 @@ char **_argv;
 
 #define _ENDIANNESS 0           /* 0 for little, 1 for big */
 
-static void longtolocal(long *l)
+static void int32_ttolocal(int32_t *l)
 {
 #if _ENDIANNESS
-    unsigned char t;
-    unsigned char *p = (unsigned char *)l;
+    uint8_t t;
+    uint8_t *p = (uint8_t *)l;
 
     t = p[0];
     p[0] = p[3];
@@ -68,6 +71,8 @@ static void longtolocal(long *l)
     t = p[1];
     p[1] = p[2];
     p[2] = p[1];
+#else
+    (void)l;             /* placate optimizers */
 #endif
 }
 
@@ -91,11 +96,11 @@ char copybytes(FILE * fp, FILE * fp2, int n)
     return (char)t;             /* return last char read */
 }
 
-long copylong(FILE * fp, FILE * fp2)
+int32_t copyint32_t(FILE * fp, FILE * fp2)
 {
-    long l;
+    int32_t l;
     int i, t;
-    unsigned char *p = (unsigned char *)&l;
+    uint8_t *p = (uint8_t *)&l;
 
     for (i = 0; i < 4; i++) {   /* skip magic no */
         t = fgetc(fp);
@@ -111,7 +116,7 @@ long copylong(FILE * fp, FILE * fp2)
             }
         *p++ = t;
     }
-    longtolocal(&l);
+    int32_ttolocal(&l);
     return l;
 }
 
@@ -120,7 +125,7 @@ int main(int argc, char **argv)
     FILE *fp, *fp2 = NULL, *fptmp;
     char *p, buf[256], c;
     int i;
-    long l;
+    int32_t l;
     time_t t;
     char rdbuf[10];
 
@@ -240,17 +245,17 @@ int main(int argc, char **argv)
                 else
                     copybytes(fp, fp2, 6);
 
-                l = copylong(fp, fp2);
+                l = copyint32_t(fp, fp2);
 
                 if (argv[1][0] == 't')
-                    printf("   %ld bytes content\n", l);
+                    printf("   %"PRId32" bytes content\n", l);
 
                 copybytes(fp, fp2, l);
             } else if ((c = copybytes(fp, fp2, 6)) >= '2') {    /* version 2 or above */
-                l = copylong(fp, fp2);
+                l = copyint32_t(fp, fp2);
 
                 if (argv[1][0] == 't')
-                    printf("RDOFF%c   %ld bytes content\n", c, l);
+                    printf("RDOFF%c   %"PRId32" bytes content\n", c, l);
                 copybytes(fp, fp2, l);  /* entire object */
             } else {
                 if (argv[1][0] == 't')
@@ -259,9 +264,9 @@ int main(int argc, char **argv)
                  * version 1 object, so we don't have an object content
                  * length field.
                  */
-                copybytes(fp, fp2, copylong(fp, fp2));  /* header */
-                copybytes(fp, fp2, copylong(fp, fp2));  /* text */
-                copybytes(fp, fp2, copylong(fp, fp2));  /* data */
+                copybytes(fp, fp2, copyint32_t(fp, fp2));  /* header */
+                copybytes(fp, fp2, copyint32_t(fp, fp2));  /* text */
+                copybytes(fp, fp2, copyint32_t(fp, fp2));  /* data */
             }
 
             if (fp2)
@@ -329,13 +334,13 @@ int main(int argc, char **argv)
             /* check against desired name */
             if (!strcmp(buf, argv[3])) {
                 fread(p = rdbuf, 1, sizeof(rdbuf), fptmp);
-                l = *(long *)(p + 6);
+                l = *(int32_t *)(p + 6);
                 fseek(fptmp, l, SEEK_CUR);
                 break;
             } else {
                 fwrite(buf, 1, strlen(buf) + 1, fp);    /* module name */
                 if ((c = copybytes(fptmp, fp, 6)) >= '2') {
-                    l = copylong(fptmp, fp);    /* version 2 or above */
+                    l = copyint32_t(fptmp, fp);    /* version 2 or above */
                     copybytes(fptmp, fp, l);    /* entire object */
                 }
             }
